/*
 * guy_init.c
 *
 *  Created on: 04 ott 2018
 *      Author: andrea
 */

#include "guy_init.h"
#include "guy.h"
#include "parameters.h"



const unsigned int scale_factor = 4;

struct _init_t{
	unsigned healthy;
	unsigned infected;
	unsigned sick;
	unsigned treatment;
	unsigned treated;
};

void guy_init(region_t *region){
	unsigned int num_healthy = args.num_healthy;
	unsigned int num_infected = args.num_infected;
	unsigned int num_sick = args.num_sick;
	unsigned int num_treatment = args.num_treatment;
	unsigned int num_treated = args.num_treated;

	(void)region;
	init_t total = {num_healthy/scale_factor, num_infected/scale_factor, num_sick/scale_factor, num_treatment/scale_factor, num_treated/scale_factor};
	init_t to_send = {0};

	const unsigned regions = RegionsCount();
	unsigned remaining = regions, i;
	rootsim_bitmap initialized[bitmap_required_size(regions)];
	bitmap_initialize(initialized, regions);


	while(remaining){
		// we pick a new random region
		i = Random()*regions;
		if(bitmap_check(initialized, i))
			continue;
		// we mark this region as initialized
		bitmap_set(initialized, i);


		// we compute the number of healthy, sick, infected etc etc people who choose this region
		to_send.healthy = random_binomial(total.healthy, 1.0/remaining);
		to_send.infected = random_binomial(total.infected, 1.0/remaining);
		to_send.sick = random_binomial(total.sick, 1.0/remaining);
		to_send.treatment = random_binomial(total.treatment, 1.0/remaining);
		to_send.treated = random_binomial(total.treated, 1.0/remaining);
		// we send the actual event
		/*printf("\n [lp %u] requested i:%u s:%u t:%u f:%u h:%u\n", i, to_send.infected, to_send.sick, to_send.treated, to_send.treatment, to_send.healthy);
		fflush(stdout);*/
		ScheduleNewEvent(i, 0.4 + 0.1*Random(), GUY_INIT, &to_send, sizeof(init_t));
		// those people just left this region
		total.healthy -= to_send.healthy;
		total.infected -= to_send.infected;
		total.sick -= to_send.sick;
		total.treatment -= to_send.treatment;
		total.treated -= to_send.treated;
		// we processed a region so we decrease the counter
		remaining--;
	}
}

static guy_t* new_base_guy(void){
	guy_t *guy_ptr = malloc(sizeof(guy_t)); //&guy_ptr e sizeof(guy_t *)
	return guy_ptr;
}

static guy_t* new_infected(void){
	guy_t *guy = new_base_guy();

	// select a random age following the given distribution
	unsigned i = 0;
	double p = Random();
	while(p_infected_age[i] < p)
		++i;

	guy->birth_day = 0 - RandomRange(365 * age_groups[i], 365 * age_groups[i+1] - 1);

	// select random origin
	if(Random() < p_infected_foreign)
		bitmap_set(guy->flags, f_foreigner);
	else
		bitmap_reset(guy->flags, f_foreigner);
	// select random gender
	if(Random() >= p_infected_male)
		bitmap_set(guy->flags, f_female);
	else
		bitmap_reset(guy->flags, f_female);

	set_risk_factors(guy);

	// set a compatible infection time xxx this is fairly different from the original model
	do{
		guy->infection_day = 0 - RandomRange(0, 365 * t_infected_max - 1);
	}while(guy->infection_day < guy->birth_day);
	// set infected state
	bitmap_reset(guy->flags, f_sick);
	bitmap_reset(guy->flags, f_treatment);

	return guy;
}

static void sickened_base_setup(guy_t *guy){
	// select a random age following the given distribution
	unsigned i = 0;
	double p = Random();
	while(p_sickened_age[i] < p)
		++i;

	guy->birth_day = 0 - RandomRange(365 * age_groups[i], 365 * age_groups[i+1] - 1);

	// select random origin
	if(Random() < p_sickened_foreign)
			bitmap_set(guy->flags, f_foreigner);
		else
			bitmap_reset(guy->flags, f_foreigner);
	// select random gender
	if(Random() >= p_sickened_male)
			bitmap_set(guy->flags, f_female);
		else
			bitmap_reset(guy->flags, f_female);
	// set smear flag
	if(Random() < p_smear)
			bitmap_set(guy->flags, f_smear);
		else
			bitmap_reset(guy->flags, f_smear);
}

static guy_t* new_sick(void){
	guy_t *guy = new_base_guy();

	sickened_base_setup(guy);

	define_diagnose(guy, 0.0);
	// we suppose this guy has already been sick for some time
	// (notice that this is equivalent to what is done in the original model
	if(guy->treatment_day)
		guy->treatment_day -= RandomRange(0, guy->treatment_day - 1);
	// set sick state
	bitmap_set(guy->flags, f_sick);
	bitmap_reset(guy->flags, f_treatment);

	return guy;
}

static guy_t* new_treatment(void){
	guy_t *guy = new_base_guy();

	sickened_base_setup(guy);
	// we suppose this guy has already been under treatment for some time
	// (notice that this is equivalent to what is done in the original model)
	guy->treatment_day = 0 - RandomRange(0, t_treatment_max - 1);
	// set treatment state
	bitmap_set(guy->flags, f_sick);
	bitmap_set(guy->flags, f_treatment);

	return guy;
}

static guy_t* new_treated(void){
	guy_t *guy = new_base_guy();

	sickened_base_setup(guy);
	// we suppose this guy has already been under treatment for some time
	// (notice that this is equivalent to what is done in the original model)
	guy->treatment_day = Random() < p_abandon ? 0 - RandomRange(t_treatment_min, t_treatment_max - 1) : 0 - (int)t_treatment_max;
	// compute relapse probability
	compute_relapse_p(guy, 0.0);
	// set treatment state
	bitmap_reset(guy->flags, f_sick);
	bitmap_set(guy->flags, f_treatment);

	return guy;
}

void guy_on_init(init_t *init_data, region_t *region){

	guy_t *infected, *sick, *treatment, *treated;
	unsigned i = init_data->infected;
	region->guys_infected = init_data->infected;


	while(i--) {
		infected = new_infected();
		try_to_insert_guy(&(region->head_infected), &(region->tail_infected), infected);
		//if (current_lp == 790) printf("[lp dest: %u] NEW INFECTED : guy %p next %p prev %p    infection_day %d\n", current_lp, infected, infected->next, infected->prev, infected->infection_day);
		scan_list_for_stats(region->head_infected, region, INSERT);
		ScheduleNewEvent(current_lp, 0.75 +  Random()/2, GUY_LEAVE, &infected, sizeof(guy_t *));
	}

	//if (current_lp == 790) printf("count infected %d count all %d\n", region->count_infected, region->count_all);

	i = init_data->sick;

	while(i--) {
		sick = new_sick();
		try_to_insert_guy(&(region->head_sick), &(region->tail_sick), sick);
		//if (current_lp == 790) printf("[lp dest: %u] NEW SICK : guy %p next %p prev %p    infection_day %d\n", current_lp, sick, sick->next, sick->prev, sick->infection_day);
		scan_list_for_stats(region->head_sick, region, INSERT);
		ScheduleNewEvent(current_lp, 0.75 +  Random()/2, GUY_LEAVE, &sick, sizeof(guy_t *));
	}

	//if (current_lp == 790) printf("count sick %d count all %d\n", region->count_sick, region->count_all);

	i = init_data->treatment;

	while(i--) {
		treatment = new_treatment();
		try_to_insert_guy(&(region->head_treatment), &(region->tail_treatment), treatment);
		//if (current_lp == 790) printf("[lp dest: %u] NEW TREAMENT : guy %p next %p prev %p    infection_day %d\n", current_lp, treatment, treatment->next, treatment->prev, treatment->infection_day);
		scan_list_for_stats(region->head_treatment, region, INSERT);
		ScheduleNewEvent(current_lp, 0.75 +  Random()/2, GUY_LEAVE, &treatment, sizeof(guy_t *));
	}

	//if (current_lp == 790) printf("count treatment %d count all %d\n", region->count_treatment, region->count_all);

	i = init_data->treated;

	while(i--) {
		treated = new_treated();
		try_to_insert_guy(&(region->head_treated), &(region->tail_treated), treated);
		//if (current_lp == 790) printf("[lp dest: %u] NEW TREATED : guy %p next %p prev %p    infection_day %d\n", current_lp, treated, treated->next, treated->prev, treated->infection_day);
		scan_list_for_stats(region->head_treated, region, INSERT);
		ScheduleNewEvent(current_lp, 0.75 +  Random()/2, GUY_LEAVE, &treated, sizeof(guy_t *));
	}

	//if (current_lp == 790) printf("count treated %d count all %d\n", region->count_treated, region->count_all);

	region->healthy = init_data->healthy;
}

